from prometheus_client import REGISTRY


METRIC_EQUALS_ERR_EXPLANATION = """
%s%s = %s, expected %s.
The values for %s are:
%s"""


class PrometheusTestCaseMixin(object):
    """A collection of utilities that make it easier to write test cases
    that interact with metrics.
    """
    def setUp(self):
        self.clearRegistry()

    def clearRegistry(self):
        """Resets the values of all collectors in the global registry.

        This is so we can test the value of exported metrics in unit
        tests.

        This is quite a hack since it relies on the internal
        representation of the prometheus_client, and it should
        probably be provided as a function there instead.
        """
        with REGISTRY._lock:
            for c in REGISTRY._collectors:
                if hasattr(c, '_metrics'):
                    c._metrics = {}
                if hasattr(c, '_value'):
                    c._value = 0.0
                if hasattr(c, '_count'):
                    c._count = 0.0
                if hasattr(c, '_sum'):
                    c._sum = 0.0
                if hasattr(c, '_buckets'):
                    c._buckets = [0.0] * len(c._buckets)

    def getMetric(self, metric_name, **labels):
        return REGISTRY.get_sample_value(metric_name, labels=labels)

    def getMetricVector(self, metric_name):
        """Returns the values for all labels of a given metric.

        The result is returned as a list of (labels, value) tuples,
        where `labels` is a dict.

        This is quite a hack since it relies on the internal
        representation of the prometheus_client, and it should
        probably be provided as a function there instead.
        """
        all_metrics = REGISTRY.collect()
        output = []
        for metric in all_metrics:
            for n, l, value in metric._samples:
                if n == metric_name:
                    output.append((l, value))
        return output

    def formatLabels(self, labels):
        """Format a set of labels to Prometheus representation.

        In:
          {'method': 'GET', 'port': '80'}

        Out:
          '{method="GET",port="80"}'
        """
        return '{%s}' % ','.join([
            '%s="%s"' % (k, v) for k, v in labels.items()])

    def formatVector(self, vector):
        """Formats a list of (labels, value) where labels is a dict into a
        human-readable representation.
        """
        return '\n'.join([
            '%s = %s' % (self.formatLabels(labels), value)
            for labels, value in vector])

    def assertMetricEquals(self, expected_value, metric_name, **labels):
        """Asserts that metric_name{**labels} == expected_value."""
        value = self.getMetric(metric_name, **labels)
        self.assertEqual(
            expected_value, value, METRIC_EQUALS_ERR_EXPLANATION % (
                metric_name, self.formatLabels(labels), value,
                expected_value, metric_name,
                self.formatVector(self.getMetricVector(metric_name))))



from django.conf.urls import patterns, url
from django_prometheus import exports

urlpatterns = patterns(
    'django_prometheus',
    url(r'^metrics$', exports.ExportToDjangoView,
        name='prometheus-django-metrics'),
)



import time
from prometheus_client import _INF

# TODO(korfuri): if python>3.3, use perf_counter() or monotonic().


def Time():
    """Returns some representation of the current time.

    This wrapper is meant to take advantage of a higher time
    resolution when available. Thus, its return value should be
    treated as an opaque object. It can be compared to the current
    time with TimeSince().
    """
    return time.time()


def TimeSince(t):
    """Compares a value returned by Time() to the current time.

    Returns:
      the time since t, in fractional seconds.
    """
    return time.time() - t


def PowersOf(logbase, count, lower=0, include_zero=True):
    """Returns a list of count powers of logbase (from logbase**lower)."""
    if not include_zero:
        return [logbase ** i for i in range(lower, count+lower)] + [_INF]
    else:
        return [0] + [logbase ** i for i in range(lower, count+lower)] + [_INF]



from django.http import HttpResponse
from django.conf import settings
import os
import prometheus_client


def SetupPrometheusEndpointOnPort(port, addr=''):
    """Exports Prometheus metrics on an HTTPServer running in its own thread.

    The server runs on the given port and is by default listenning on
    all interfaces. This HTTPServer is fully independent of Django and
    its stack. This offers the advantage that even if Django becomes
    unable to respond, the HTTPServer will continue to function and
    export metrics. However, this also means that none of the features
    offered by Django (like middlewares or WSGI) can't be used.

    Now here's the really weird part. When Django runs with the
    auto-reloader enabled (which is the default, you can disable it
    with `manage.py runserver --noreload`), it forks and executes
    manage.py twice. That's wasteful but usually OK. It starts being a
    problem when you try to open a port, like we do. We can detect
    that we're running under an autoreloader through the presence of
    the RUN_MAIN environment variable, so we abort if we're trying to
    export under an autoreloader and trying to open a port.
    """
    assert os.environ.get('RUN_MAIN') != 'true', (
        'The thread-based exporter can\'t be safely used when django\'s '
        'autoreloader is active. Use the URL exporter, or start django '
        'with --noreload. See documentation/exports.md.')
    prometheus_client.start_http_server(port, addr=addr)


def SetupPrometheusExportsFromConfig():
    """Exports metrics so Prometheus can collect them."""
    port = getattr(settings, 'PROMETHEUS_METRICS_EXPORT_PORT', None)
    addr = getattr(settings, 'PROMETHEUS_METRICS_EXPORT_ADDRESS', None)
    if port:
        SetupPrometheusEndpointOnPort(port, addr)


def ExportToDjangoView(request):
    """Exports /metrics as a Django view.

    You can use django_prometheus.urls to map /metrics to this view.
    """
    metrics_page = prometheus_client.generate_latest()
    return HttpResponse(
        metrics_page,
        content_type=prometheus_client.CONTENT_TYPE_LATEST)



from django.apps import AppConfig
from django_prometheus.exports import SetupPrometheusExportsFromConfig
# unused import to force instantiating the metric objects at startup.
import django_prometheus


class DjangoPrometheusConfig(AppConfig):
    name = 'django_prometheus'
    verbose_name = 'Django-Prometheus'

    def ready(self):
        """Initializes the Prometheus exports if they are enabled in the config.

        Note that this is called even for other management commands
        than `runserver`. As such, it is possible to scrape the
        metrics of a running `manage.py test` or of another command,
        which shouldn't be done for real monitoring (since these jobs
        are usually short-lived), but can be useful for debugging.
        """
        SetupPrometheusExportsFromConfig()



from prometheus_client import Counter, Histogram
from django_prometheus.utils import Time, TimeSince, PowersOf

requests_total = Counter(
    'django_http_requests_before_middlewares_total',
    'Total count of requests before middlewares run.')
responses_total = Counter(
    'django_http_responses_before_middlewares_total',
    'Total count of responses before middlewares run.')
requests_latency_before = Histogram(
    'django_http_requests_latency_including_middlewares_seconds',
    ('Histogram of requests processing time (including middleware '
     'processing time).'))
requests_unknown_latency_before = Counter(
    'django_http_requests_unknown_latency_including_middlewares_total',
    ('Count of requests for which the latency was unknown (when computing '
     'django_http_requests_latency_including_middlewares_seconds).'))


class PrometheusBeforeMiddleware(object):
    """Monitoring middleware that should run before other middlewares."""
    def process_request(self, request):
        requests_total.inc()
        request.prometheus_before_middleware_event = Time()

    def process_response(self, request, response):
        responses_total.inc()
        if hasattr(request, 'prometheus_before_middleware_event'):
            requests_latency_before.observe(TimeSince(
                request.prometheus_before_middleware_event))
        else:
            requests_unknown_latency_before.inc()
        return response


requests_latency = Histogram(
    'django_http_requests_latency_seconds',
    'Histogram of requests processing time.')
requests_unknown_latency = Counter(
    'django_http_requests_unknown_latency_total',
    'Count of requests for which the latency was unknown.')
# Set in process_request
ajax_requests = Counter(
    'django_http_ajax_requests_total',
    'Count of AJAX requests.')
requests_by_method = Counter(
    'django_http_requests_total_by_method',
    'Count of requests by method.',
    ['method'])
requests_by_transport = Counter(
    'django_http_requests_total_by_transport',
    'Count of requests by transport.',
    ['transport'])
# Set in process_view
requests_by_view_transport_method = Counter(
    'django_http_requests_total_by_view_transport_method',
    'Count of requests by view, transport, method.',
    ['view', 'transport', 'method'])
requests_body_bytes = Histogram(
    'django_http_requests_body_total_bytes',
    'Histogram of requests by body size.',
    buckets=PowersOf(2, 30))
# Set in process_template_response
responses_by_templatename = Counter(
    'django_http_responses_total_by_templatename',
    'Count of responses by template name.',
    ['templatename'])
# Set in process_response
responses_by_status = Counter(
    'django_http_responses_total_by_status',
    'Count of responses by status.',
    ['status'])
responses_body_bytes = Histogram(
    'django_http_responses_body_total_bytes',
    'Histogram of responses by body size.',
    buckets=PowersOf(2, 30))
responses_by_charset = Counter(
    'django_http_responses_total_by_charset',
    'Count of responses by charset.',
    ['charset'])
responses_streaming = Counter(
    'django_http_responses_streaming_total',
    'Count of streaming responses.')
# Set in process_exception
exceptions_by_type = Counter(
    'django_http_exceptions_total_by_type',
    'Count of exceptions by object type.',
    ['type'])
exceptions_by_view = Counter(
    'django_http_exceptions_total_by_view',
    'Count of exceptions by view.',
    ['view_name'])


class PrometheusAfterMiddleware(object):
    """Monitoring middleware that should run after other middlewares."""
    def _transport(self, request):
        return 'https' if request.is_secure() else 'http'

    def _method(self, request):
        m = request.method
        if m not in ('GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'TRACE',
                     'OPTIONS', 'CONNECT', 'PATCH'):
            return '<invalid method>'
        return m

    def process_request(self, request):
        transport = self._transport(request)
        method = self._method(request)
        requests_by_method.labels(method).inc()
        requests_by_transport.labels(transport).inc()
        if request.is_ajax():
            ajax_requests.inc()
        requests_body_bytes.observe(len(request.body))
        request.prometheus_after_middleware_event = Time()

    def process_view(self, request, view_func, *view_args, **view_kwargs):
        transport = self._transport(request)
        method = self._method(request)
        name = request.resolver_match.view_name or '<unnamed view>'
        requests_by_view_transport_method.labels(
            name, transport, method).inc()

    def process_template_response(self, request, response):
        responses_by_templatename.labels(str(
            response.template_name)).inc()
        return response

    def process_response(self, request, response):
        responses_by_status.labels(str(response.status_code)).inc()
        if hasattr(response, 'charset'):
            responses_by_charset.labels(str(response.charset)).inc()
        if response.streaming:
            responses_streaming.inc()
        responses_body_bytes.observe(len(response.content))
        if hasattr(request, 'prometheus_after_middleware_event'):
            requests_latency.observe(TimeSince(
                request.prometheus_after_middleware_event))
        else:
            requests_unknown_latency.inc()
        return response

    def process_exception(self, request, exception):
        name = request.resolver_match.view_name or '<unnamed view>'
        exceptions_by_type.labels(type(exception).__name__).inc()
        exceptions_by_view.labels(name).inc()
        if hasattr(request, 'prometheus_after_middleware_event'):
            requests_latency.observe(TimeSince(
                request.prometheus_after_middleware_event))
        else:
            requests_unknown_latency.inc()
